//! Response models for the Bandcamp API

use std::collections::BTreeMap;

use serde::{Deserialize, Serialize};
use serde_with::{serde_as, NoneAsEmptyString, VecSkipError};
use time::OffsetDateTime;

use crate::util;

/// Bandcamp band object
#[serde_as]
#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
#[non_exhaustive]
pub struct Band {
    /// Numeric band ID
    pub id: u64,
    /// Band name
    pub name: String,
    /// Bandcamp URL
    ///
    /// **Example:** `https://wintergatan.bandcamp.com/`
    pub bandcamp_url: String,
    /// Band biography text
    #[serde_as(deserialize_as = "NoneAsEmptyString")]
    pub bio: Option<String>,
    /// Avatar image ID
    ///
    /// Image URL: `https://f4.bcbits.com/img/<ID>_<SIZE>.jpg`
    pub bio_image_id: Option<u64>,
    /// Header image ID
    ///
    /// Image URL: `https://f4.bcbits.com/img/<ID>_<SIZE>.jpg`
    pub header_image_id: Option<u64>,
    /// Band location (<City>, <Country>)
    #[serde_as(deserialize_as = "NoneAsEmptyString")]
    pub location: Option<String>,
    /// List of links to other websites or social media profiles
    pub sites: Vec<Site>,
    /// List of albums and single tracks of the band
    pub discography: Vec<Discograph>,
}

/// Bandcamp album object
#[serde_as]
#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
#[non_exhaustive]
pub struct Album {
    /// Numeric album ID
    pub id: u64,
    /// Album title
    pub title: String,
    /// Album art ID
    ///
    /// Image URL: `https://f4.bcbits.com/img/a<ID>_<SIZE>.jpg`
    pub art_id: Option<u64>,
    /// Album type: track (single) or album
    #[serde(rename = "type")]
    pub tralbum_type: TrAlbumType,
    /// Album artist name
    pub tralbum_artist: String,
    /// Album band
    pub band: BandInfo,
    /// Bandcamp url of the album
    ///
    /// **Example:** `https://wintergatan.bandcamp.com/album/wintergatan-full-album`
    pub bandcamp_url: String,
    /// Album credits
    #[serde_as(deserialize_as = "NoneAsEmptyString")]
    pub credits: Option<String>,
    /// Album description text
    #[serde_as(deserialize_as = "NoneAsEmptyString")]
    pub about: Option<String>,
    /// ID of the featured track (displayed at the top of the page)
    pub featured_track_id: u64,
    /// Album release date
    #[serde(with = "time::serde::timestamp")]
    pub release_date: OffsetDateTime,
    /// List of album tags (genres and locations)
    #[serde(default)]
    pub tags: Vec<Tag>,
    /// List of album tracks
    pub tracks: Vec<Track>,
}

/// Bandcamp track object
#[serde_as]
#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
#[non_exhaustive]
pub struct Track {
    /// Numeric track ID
    pub track_id: u64,
    /// Track title
    pub title: String,
    /// Numeric album id
    pub album_id: u64,
    /// Album title
    #[serde_as(deserialize_as = "NoneAsEmptyString")]
    pub album_title: Option<String>,
    /// Numeric band id
    pub band_id: u64,
    /// Band name
    #[serde_as(deserialize_as = "NoneAsEmptyString")]
    pub band_name: Option<String>,
    /// Duration in seconds
    pub duration: f32,
    /// Map of audio formats and URLs.
    ///
    /// 128kbit/s MP3 is currently the only publicly available format.
    ///
    /// **Example:**
    ///
    /// `mp3-128` => `https://bandcamp.com/stream_redirect?enc=mp3-128&track_id=4173325157&ts=1686871135&t=a1f7904205e6caca970009ca6744ef0b512f353a`
    #[serde(default)]
    pub streaming_url: BTreeMap<String, String>,
    /// Album track number
    pub track_num: u16,
}

/// Discograph (album or track)
#[serde_as]
#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
#[non_exhaustive]
pub struct Discograph {
    /// Numeric track/album ID
    pub item_id: u64,
    /// Item type (album/track)
    pub item_type: DiscographType,
    /// Album/Track title
    pub title: String,
    /// Album art ID
    ///
    /// Image URL: `https://f4.bcbits.com/img/a<ID>_<SIZE>.jpg`
    pub art_id: Option<u64>,
    /// Artist name (if different from band name)
    #[serde_as(deserialize_as = "NoneAsEmptyString")]
    pub artist_name: Option<String>,
    /// Numeric band ID
    pub band_id: u64,
    /// Band name
    pub band_name: String,
    /// Release date
    #[serde(with = "crate::serializer::date")]
    pub release_date: OffsetDateTime,
}

/// Album type of a discograph
#[derive(Debug, Copy, Clone, PartialEq, Eq, Serialize, Deserialize)]
#[serde(rename_all = "snake_case")]
pub enum DiscographType {
    /// Album
    Album,
    /// Track (single)
    Track,
}

/// Band information
#[serde_as]
#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
#[non_exhaustive]
pub struct BandInfo {
    /// Numeric band ID
    pub band_id: u64,
    /// Band name
    pub name: String,
    /// Band biography text
    #[serde_as(deserialize_as = "NoneAsEmptyString")]
    pub bio: Option<String>,
    /// Avatar image ID
    ///
    /// Image URL: `https://f4.bcbits.com/img/<ID>_<SIZE>.jpg`
    pub image_id: Option<u64>,
    /// Band location (<City>, <Country>)
    #[serde_as(deserialize_as = "NoneAsEmptyString")]
    pub location: Option<String>,
}

/// Album tag
#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
#[non_exhaustive]
pub struct Tag {
    /// Displayed tag name
    ///
    /// **Example:** `punk rock`
    pub name: String,
    /// Tag name slug
    ///
    /// **Example:** `punk-rock`
    pub norm_name: String,
    /// True if the tag describes a location
    pub isloc: bool,
}

/// Link to an external website or social media profile
#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
#[non_exhaustive]
pub struct Site {
    /// Website title
    pub title: String,
    /// Website URL
    pub url: String,
}

/// Search result
#[serde_as]
#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
#[non_exhaustive]
pub struct SearchResult {
    /// List of found items
    #[serde_as(as = "VecSkipError<_>")]
    pub results: Vec<SearchItem>,
    /// Time in milliseconds spent on the search
    pub time_ms: u32,
}

/// Search item
#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
#[non_exhaustive]
#[allow(missing_docs)]
#[serde(tag = "type")]
pub enum SearchItem {
    #[serde(rename = "b")]
    Band(SearchBand),
    #[serde(rename = "a")]
    Album(SearchAlbum),
    #[serde(rename = "t")]
    Track(SearchTrack),
}

/// Band from search results
#[serde_as]
#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
#[non_exhaustive]
pub struct SearchBand {
    /// Numeric band ID
    pub id: u64,
    /// Band name
    pub name: String,
    /// Avatar image ID
    ///
    /// Image URL: `https://f4.bcbits.com/img/<ID>_<SIZE>.jpg`
    pub img_id: Option<u64>,
    /// Band location (<City>, <Country>)
    #[serde_as(deserialize_as = "NoneAsEmptyString")]
    pub location: Option<String>,
    /// List of tag names associated with the
    #[serde(default)]
    pub tag_names: Vec<String>,
    /// Bandcamp URL
    ///
    /// **Example:** `https://wintergatan.bandcamp.com/`
    pub url: String,
}

/// Album from search results
#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
#[non_exhaustive]
pub struct SearchAlbum {
    /// Numeric album ID
    pub id: u64,
    /// Album title
    pub name: String,
    /// Numeric band ID
    pub band_id: u64,
    /// Band name
    pub band_name: String,
    /// Album art ID
    ///
    /// Image URL: `https://f4.bcbits.com/img/a<ID>_<SIZE>.jpg`
    pub art_id: Option<u64>,
}

/// Track from search results
#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
#[non_exhaustive]
pub struct SearchTrack {
    /// Numeric track ID
    pub id: u64,
    /// Track title
    pub name: String,
    /// Numeric band ID
    pub band_id: u64,
    /// Band name
    pub band_name: String,
    /// Album art ID
    ///
    /// Image URL: `https://f4.bcbits.com/img/a<ID>_<SIZE>.jpg`
    pub art_id: Option<u64>,
    /// Numeric album ID
    pub album_id: Option<u64>,
    /// Album name
    pub album_name: Option<String>,
}

/// Bandcamp feed containing featured items
#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
#[non_exhaustive]
pub struct Feed {
    /// Update date of the feed
    #[serde(with = "time::serde::timestamp")]
    pub feed_timestamp: OffsetDateTime,
    /// List of feed entries
    pub stories: FeedStories,
}

/// List of feed entries
#[serde_as]
#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
#[non_exhaustive]
pub struct FeedStories {
    /// List of feed entries
    #[serde_as(as = "VecSkipError<_>")]
    pub featured: Vec<FeedItem>,
}

/// Bandcamp feed item
#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
#[non_exhaustive]
pub struct FeedItem {
    /// Numeric track/album ID
    pub tralbum_id: u64,
    /// Album type: track (single) or album
    pub tralbum_type: TrAlbumType,
    /// Title of the feed item
    pub title: String,
    /// Artist name
    pub artist: Option<String>,
    /// Numeric band ID
    pub band_id: u64,
    /// Band name
    pub band_name: String,
    /// Bandcamp URL of the band
    ///
    /// **Example:** `https://wintergatan.bandcamp.com/`
    pub band_url: String,
    /// Album art ID
    ///
    /// Image URL: `https://f4.bcbits.com/img/a<ID>_<SIZE>.jpg`
    pub art_id: Option<u64>,
    /// Feed item ID
    pub ntid: u64,
    /// Feed item date
    #[serde(with = "time::serde::timestamp")]
    pub story_date: OffsetDateTime,
    /// Feed item type (Default/Ad)
    pub story_type: StoryType,
    /// Used to fetch continuations
    pub story_token: String,
    /// Text of the feed item
    #[serde(alias = "description")]
    pub desc: String,
    /// Full text of the feed item
    pub full_text: Option<String>,
    /// Some story entries link to an article on daily.bandcamp.com
    pub article_url: Option<String>,
    /// Number of tracks of the album
    pub num_streamable_tracks: u16,
    /// List of tags
    #[serde(default)]
    pub tags: Vec<Tag>,
}

/// Feed story type
#[derive(Debug, Copy, Clone, PartialEq, Eq, Serialize, Deserialize)]
#[non_exhaustive]
#[allow(missing_docs)]
pub enum StoryType {
    #[serde(rename = "nn")]
    Default,
    #[serde(rename = "ad")]
    Ad,
}

/// Track/Album type
#[derive(Debug, Copy, Clone, PartialEq, Eq, Serialize, Deserialize)]
pub enum TrAlbumType {
    /// Album
    #[serde(rename = "a")]
    Album,
    /// Track (Single)
    #[serde(rename = "t")]
    Track,
}

impl TrAlbumType {
    pub(crate) fn char(&self) -> char {
        match self {
            TrAlbumType::Album => 'a',
            TrAlbumType::Track => 't',
        }
    }
}

impl From<DiscographType> for TrAlbumType {
    fn from(value: DiscographType) -> Self {
        match value {
            DiscographType::Album => Self::Album,
            DiscographType::Track => Self::Track,
        }
    }
}

/// Search filter
#[derive(Debug, Copy, Clone, PartialEq, Eq, Serialize, Deserialize)]
#[non_exhaustive]
#[allow(missing_docs)]
pub enum SearchFilter {
    #[serde(rename = "")]
    All,
    #[serde(rename = "b")]
    Bands,
    #[serde(rename = "a")]
    Albums,
    #[serde(rename = "t")]
    Tracks,
}

// Album UID

impl Album {
    /// Get a unified album ID.
    ///
    /// Unified album IDs allow for fetching Bandcamp albums using one string
    /// parameter. They are not an official part of Bandcamp.
    pub fn album_uid(&self) -> String {
        util::create_album_uid(self.band.band_id, self.tralbum_type, self.id)
    }
}

impl Discograph {
    /// Get a unified album ID.
    ///
    /// Unified album IDs allow for fetching Bandcamp albums using one string
    /// parameter. They are not an official part of Bandcamp.
    pub fn album_uid(&self) -> String {
        util::create_album_uid(self.band_id, self.item_type.into(), self.item_id)
    }
}

impl SearchAlbum {
    /// Get a unified album ID.
    ///
    /// Unified album IDs allow for fetching Bandcamp albums using one string
    /// parameter. They are not an official part of Bandcamp.
    pub fn album_uid(&self) -> String {
        util::create_album_uid(self.band_id, TrAlbumType::Album, self.id)
    }
}

impl FeedItem {
    /// Get a unified album ID.
    ///
    /// Unified album IDs allow for fetching Bandcamp albums using one string
    /// parameter. They are not an official part of Bandcamp.
    pub fn album_uid(&self) -> String {
        util::create_album_uid(self.band_id, self.tralbum_type, self.tralbum_id)
    }
}
