use serde::{de::Visitor, Deserializer, Serialize, Serializer};

pub mod date {
    use super::*;
    use time::{macros::format_description, OffsetDateTime, PrimitiveDateTime};

    const DATE_FORMAT: &[time::format_description::FormatItem] =
        format_description!("[day] [month repr:short] [year] [hour]:[minute]:[second] GMT");

    pub fn serialize<S: Serializer>(
        value: &OffsetDateTime,
        serializer: S,
    ) -> Result<S::Ok, S::Error> {
        value
            .format(&DATE_FORMAT)
            .map_err(serde::ser::Error::custom)?
            .serialize(serializer)
    }

    pub fn deserialize<'a, D: Deserializer<'a>>(
        deserializer: D,
    ) -> Result<OffsetDateTime, D::Error> {
        struct DateVisitor;

        impl<'de> Visitor<'de> for DateVisitor {
            type Value = OffsetDateTime;

            fn expecting(&self, formatter: &mut std::fmt::Formatter) -> std::fmt::Result {
                formatter.write_str("datetime in the format of `24 Apr 2013 08:30:00 GMT`")
            }

            fn visit_str<E>(self, v: &str) -> Result<Self::Value, E>
            where
                E: serde::de::Error,
            {
                PrimitiveDateTime::parse(v, &DATE_FORMAT)
                    .map(PrimitiveDateTime::assume_utc)
                    .map_err(serde::de::Error::custom)
            }
        }

        deserializer.deserialize_str(DateVisitor)
    }
}

#[cfg(test)]
mod tests {
    use serde::Deserialize;
    use time::{macros::datetime, OffsetDateTime};

    use super::*;

    #[test]
    fn deserialize_date() {
        #[derive(Deserialize)]
        struct S {
            #[serde(with = "date")]
            d: OffsetDateTime,
        }

        let json = r#"{"d": "21 Apr 2023 08:30:00 GMT"}"#;
        assert_eq!(
            serde_json::from_str::<S>(json).unwrap().d,
            datetime!(2023-04-21 8:30 +0)
        );
    }
}
