use crate::{error::Result, models::TrAlbumType, Error};

pub fn create_album_uid(band_id: u64, tralbum_type: TrAlbumType, album_id: u64) -> String {
    format!("{}{}{}", band_id, tralbum_type.char(), album_id)
}

pub fn parse_album_uid<S: AsRef<str>>(uid: S) -> Result<(u64, TrAlbumType, u64)> {
    let uid = uid.as_ref();
    let atype_i = uid.find(['a', 't']).ok_or(Error::InvalidUid)?;

    let band_id = uid[0..atype_i].parse().map_err(|_| Error::InvalidUid)?;
    let tralbum_type = match &uid[atype_i..atype_i + 1] {
        "a" => TrAlbumType::Album,
        "t" => TrAlbumType::Track,
        _ => return Err(Error::InvalidUid),
    };
    let album_id = uid[atype_i + 1..].parse().map_err(|_| Error::InvalidUid)?;

    Ok((band_id, tralbum_type, album_id))
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn t_create_album_uid() {
        assert_eq!(create_album_uid(123, TrAlbumType::Track, 456), "123t456");
        assert_eq!(
            create_album_uid(123, TrAlbumType::Album, 456789),
            "123a456789"
        );
    }

    #[test]
    fn t_parse_album_uid() {
        assert_eq!(
            parse_album_uid("123t456").unwrap(),
            (123, TrAlbumType::Track, 456)
        );
        assert_eq!(
            parse_album_uid("123a456789").unwrap(),
            (123, TrAlbumType::Album, 456789)
        );
    }

    #[test]
    fn t_parse_album_uid_error() {
        let bad_uids = &["", "a", "123", "123a", "a123", "123x123", "x123a123"];

        bad_uids.iter().for_each(|uid| {
            let err = parse_album_uid(uid);
            assert!(matches!(err, Err(Error::InvalidUid)), "uid: {uid}");
        });
    }
}
